@@ -16,7 +16,6 @@
 // not hash_map: it's not as portable and needs hash<string>.
 #include <map>                          // for map, map<>::mapped_type, etc
 #include <memory>
-#include <regex>
 #include <string>                       // for string, basic_string, etc
 #include <system_error>                 // for error_code
 #include <utility>                      // for pair, make_pair
@@ -25,6 +24,7 @@
 #include "iwyu_location_util.h"
 #include "iwyu_path_util.h"
 #include "iwyu_port.h"
+#include "iwyu_regex.h"
 #include "iwyu_stl_util.h"
 #include "iwyu_string_util.h"
 #include "iwyu_verrs.h"
@@ -37,6 +37,7 @@
 #include "llvm/Support/SourceMgr.h"
 #include "llvm/Support/YAMLParser.h"
 #include "clang/Basic/FileManager.h"
+#include "clang/Tooling/Inclusions/StandardLibrary.h"
 
 using std::find;
 using std::make_pair;
@@ -91,53 +92,38 @@ const IncludeMapEntry libc_symbol_map[] = {
   // we always set it to kPrivate.
   { "aiocb", kPrivate, "<aio.h>", kPublic },
   { "blkcnt_t", kPrivate, "<sys/types.h>", kPublic },
-  { "blkcnt_t", kPrivate, "<sys/stat.h>", kPublic },
   { "blksize_t", kPrivate, "<sys/types.h>", kPublic },
-  { "blksize_t", kPrivate, "<sys/stat.h>", kPublic },
   { "cc_t", kPrivate, "<termios.h>", kPublic },
-  { "clock_t", kPrivate, "<sys/types.h>", kPublic },
-  { "clock_t", kPrivate, "<sys/time.h>", kPublic },
   { "clock_t", kPrivate, "<time.h>", kPublic },
+  { "clock_t", kPrivate, "<sys/types.h>", kPublic },
   { "clockid_t", kPrivate, "<sys/types.h>", kPublic },
-  { "clockid_t", kPrivate, "<time.h>", kPublic },
   { "daddr_t", kPrivate, "<sys/types.h>", kPublic },
-  { "daddr_t", kPrivate, "<rpc/types.h>", kPublic },
   { "dev_t", kPrivate, "<sys/types.h>", kPublic },
-  { "dev_t", kPrivate, "<sys/stat.h>", kPublic },
   { "div_t", kPrivate, "<stdlib.h>", kPublic },
   { "double_t", kPrivate, "<math.h>", kPublic },
   { "error_t", kPrivate, "<errno.h>", kPublic },
   { "error_t", kPrivate, "<argp.h>", kPublic },
   { "error_t", kPrivate, "<argz.h>", kPublic },
   { "fd_set", kPrivate, "<sys/select.h>", kPublic },
-  { "fd_set", kPrivate, "<sys/time.h>", kPublic },
   { "fenv_t", kPrivate, "<fenv.h>", kPublic },
   { "fexcept_t", kPrivate, "<fenv.h>", kPublic },
   { "FILE", kPrivate, "<stdio.h>", kPublic },
   { "FILE", kPrivate, "<wchar.h>", kPublic },
   { "float_t", kPrivate, "<math.h>", kPublic },
   { "fsblkcnt_t", kPrivate, "<sys/types.h>", kPublic },
-  { "fsblkcnt_t", kPrivate, "<sys/statvfs.h>", kPublic },
   { "fsfilcnt_t", kPrivate, "<sys/types.h>", kPublic },
-  { "fsfilcnt_t", kPrivate, "<sys/statvfs.h>", kPublic },
+  { "getopt", kPrivate, "<unistd.h>", kPublic },
   { "gid_t", kPrivate, "<sys/types.h>", kPublic },
-  { "gid_t", kPrivate, "<grp.h>", kPublic },
-  { "gid_t", kPrivate, "<pwd.h>", kPublic },
-  { "gid_t", kPrivate, "<signal.h>", kPublic },
-  { "gid_t", kPrivate, "<stropts.h>", kPublic },
-  { "gid_t", kPrivate, "<sys/ipc.h>", kPublic },
-  { "gid_t", kPrivate, "<sys/stat.h>", kPublic },
   { "gid_t", kPrivate, "<unistd.h>", kPublic },
+  { "htonl", kPrivate, "<arpa/inet.h>", kPublic },
+  { "htons", kPrivate, "<arpa/inet.h>", kPublic },
   { "id_t", kPrivate, "<sys/types.h>", kPublic },
-  { "id_t", kPrivate, "<sys/resource.h>", kPublic },
   { "imaxdiv_t", kPrivate, "<inttypes.h>", kPublic },
   { "intmax_t", kPrivate, "<stdint.h>", kPublic },
   { "uintmax_t", kPrivate, "<stdint.h>", kPublic },
   { "ino64_t", kPrivate, "<sys/types.h>", kPublic },
   { "ino64_t", kPrivate, "<dirent.h>", kPublic },
   { "ino_t", kPrivate, "<sys/types.h>", kPublic },
-  { "ino_t", kPrivate, "<dirent.h>", kPublic },
-  { "ino_t", kPrivate, "<sys/stat.h>", kPublic },
   { "int8_t", kPrivate, "<stdint.h>", kPublic },
   { "int16_t", kPrivate, "<stdint.h>", kPublic },
   { "int32_t", kPrivate, "<stdint.h>", kPublic },
@@ -148,105 +134,50 @@ const IncludeMapEntry libc_symbol_map[] = {
   { "uint64_t", kPrivate, "<stdint.h>", kPublic },
   { "intptr_t", kPrivate, "<stdint.h>", kPublic },
   { "uintptr_t", kPrivate, "<stdint.h>", kPublic },
+  { "iovec", kPrivate, "<sys/uio.h>", kPublic },
+  { "itimerspec", kPrivate, "<time.h>", kPublic },
   { "key_t", kPrivate, "<sys/types.h>", kPublic },
-  { "key_t", kPrivate, "<sys/ipc.h>", kPublic },
   { "lconv", kPrivate, "<locale.h>", kPublic },
   { "ldiv_t", kPrivate, "<stdlib.h>", kPublic },
   { "lldiv_t", kPrivate, "<stdlib.h>", kPublic },
   { "max_align_t", kPrivate, "<stddef.h>", kPublic },
   { "mode_t", kPrivate, "<sys/types.h>", kPublic },
-  { "mode_t", kPrivate, "<fcntl.h>", kPublic },
-  { "mode_t", kPrivate, "<ndbm.h>", kPublic },
-  { "mode_t", kPrivate, "<spawn.h>", kPublic },
-  { "mode_t", kPrivate, "<sys/ipc.h>", kPublic },
-  { "mode_t", kPrivate, "<sys/mman.h>", kPublic },
-  { "mode_t", kPrivate, "<sys/stat.h>", kPublic },
   { "nlink_t", kPrivate, "<sys/types.h>", kPublic },
-  { "nlink_t", kPrivate, "<sys/stat.h>", kPublic },
+  { "ntohl", kPrivate, "<arpa/inet.h>", kPublic },
+  { "ntohs", kPrivate, "<arpa/inet.h>", kPublic },
   { "off64_t", kPrivate, "<sys/types.h>", kPublic },
   { "off64_t", kPrivate, "<unistd.h>", kPublic },
   { "off_t", kPrivate, "<sys/types.h>", kPublic },
-  { "off_t", kPrivate, "<aio.h>", kPublic },
-  { "off_t", kPrivate, "<fcntl.h>", kPublic },
-  { "off_t", kPrivate, "<stdio.h>", kPublic },
-  { "off_t", kPrivate, "<sys/mman.h>", kPublic },
-  { "off_t", kPrivate, "<sys/stat.h>", kPublic },
   { "off_t", kPrivate, "<unistd.h>", kPublic },
+  { "optarg", kPrivate, "<unistd.h>", kPublic },
+  { "opterr", kPrivate, "<unistd.h>", kPublic },
+  { "optind", kPrivate, "<unistd.h>", kPublic },
+  { "optopt", kPrivate, "<unistd.h>", kPublic },
   { "pid_t", kPrivate, "<sys/types.h>", kPublic },
-  { "pid_t", kPrivate, "<fcntl.h>", kPublic },
-  { "pid_t", kPrivate, "<sched.h>", kPublic },
-  { "pid_t", kPrivate, "<signal.h>", kPublic },
-  { "pid_t", kPrivate, "<spawn.h>", kPublic },
-  { "pid_t", kPrivate, "<sys/msg.h>", kPublic },
-  { "pid_t", kPrivate, "<sys/sem.h>", kPublic },
-  { "pid_t", kPrivate, "<sys/shm.h>", kPublic },
-  { "pid_t", kPrivate, "<sys/wait.h>", kPublic },
-  { "pid_t", kPrivate, "<termios.h>", kPublic },
-  { "pid_t", kPrivate, "<time.h>", kPublic },
   { "pid_t", kPrivate, "<unistd.h>", kPublic },
-  { "pid_t", kPrivate, "<utmpx.h>", kPublic },
   { "ptrdiff_t", kPrivate, "<stddef.h>", kPublic },
   { "regex_t", kPrivate, "<regex.h>", kPublic },
   { "regmatch_t", kPrivate, "<regex.h>", kPublic },
   { "regoff_t", kPrivate, "<regex.h>", kPublic },
   { "sigevent", kPrivate, "<signal.h>", kPublic },
-  { "sigevent", kPrivate, "<aio.h>", kPublic },
-  { "sigevent", kPrivate, "<mqueue.h>", kPublic },
-  { "sigevent", kPrivate, "<time.h>", kPublic },
   { "siginfo_t", kPrivate, "<signal.h>", kPublic },
-  { "siginfo_t", kPrivate, "<sys/wait.h>", kPublic },
   { "sigset_t", kPrivate, "<signal.h>", kPublic },
-  { "sigset_t", kPrivate, "<spawn.h>", kPublic },
-  { "sigset_t", kPrivate, "<sys/select.h>", kPublic },
   { "sigval", kPrivate, "<signal.h>", kPublic },
   { "sockaddr", kPrivate, "<sys/socket.h>", kPublic },
   { "socklen_t", kPrivate, "<sys/socket.h>", kPublic },
-  { "socklen_t", kPrivate, "<netdb.h>", kPublic },
   { "ssize_t", kPrivate, "<sys/types.h>", kPublic },
-  { "ssize_t", kPrivate, "<aio.h>", kPublic },
-  { "ssize_t", kPrivate, "<monetary.h>", kPublic },
-  { "ssize_t", kPrivate, "<mqueue.h>", kPublic },
-  { "ssize_t", kPrivate, "<stdio.h>", kPublic },
-  { "ssize_t", kPrivate, "<sys/msg.h>", kPublic },
-  { "ssize_t", kPrivate, "<sys/socket.h>", kPublic },
-  { "ssize_t", kPrivate, "<sys/uio.h>", kPublic },
   { "ssize_t", kPrivate, "<unistd.h>", kPublic },
   { "stat", kPrivate, "<sys/stat.h>", kPublic },
-  { "stat", kPrivate, "<ftw.h>", kPublic },
   { "suseconds_t", kPrivate, "<sys/types.h>", kPublic },
-  { "suseconds_t", kPrivate, "<sys/select.h>", kPublic },
-  { "suseconds_t", kPrivate, "<sys/time.h>", kPublic },
   { "time_t", kPrivate, "<time.h>", kPublic },
-  { "time_t", kPrivate, "<sched.h>", kPublic },
-  { "time_t", kPrivate, "<sys/msg.h>", kPublic },
-  { "time_t", kPrivate, "<sys/select.h>", kPublic },
-  { "time_t", kPrivate, "<sys/sem.h>", kPublic },
-  { "time_t", kPrivate, "<sys/shm.h>", kPublic },
-  { "time_t", kPrivate, "<sys/stat.h>", kPublic },
-  { "time_t", kPrivate, "<sys/time.h>", kPublic },
   { "time_t", kPrivate, "<sys/types.h>", kPublic },
-  { "time_t", kPrivate, "<utime.h>", kPublic },
   { "timer_t", kPrivate, "<sys/types.h>", kPublic },
-  { "timer_t", kPrivate, "<time.h>", kPublic },
   { "timespec", kPrivate, "<time.h>", kPublic },
-  { "timespec", kPrivate, "<aio.h>", kPublic },
-  { "timespec", kPrivate, "<mqueue.h>", kPublic },
-  { "timespec", kPrivate, "<sched.h>", kPublic },
   { "timespec", kPrivate, "<signal.h>", kPublic },
-  { "timespec", kPrivate, "<sys/select.h>", kPublic },
-  { "timespec", kPrivate, "<sys/stat.h>", kPublic },
   { "timeval", kPrivate, "<sys/time.h>", kPublic },
-  { "timeval", kPrivate, "<sys/resource.h>", kPublic },
-  { "timeval", kPrivate, "<sys/select.h>", kPublic },
-  { "timeval", kPrivate, "<utmpx.h>", kPublic },
+  { "tm", kPrivate, "<time.h>", kPublic },
   { "u_char", kPrivate, "<sys/types.h>", kPublic },
-  { "u_char", kPrivate, "<rpc/types.h>", kPublic },
   { "uid_t", kPrivate, "<sys/types.h>", kPublic },
-  { "uid_t", kPrivate, "<pwd.h>", kPublic },
-  { "uid_t", kPrivate, "<signal.h>", kPublic },
-  { "uid_t", kPrivate, "<stropts.h>", kPublic },
-  { "uid_t", kPrivate, "<sys/ipc.h>", kPublic },
-  { "uid_t", kPrivate, "<sys/stat.h>", kPublic },
   { "uid_t", kPrivate, "<unistd.h>", kPublic },
   { "useconds_t", kPrivate, "<sys/types.h>", kPublic },
   { "useconds_t", kPrivate, "<unistd.h>", kPublic },
@@ -255,44 +186,30 @@ const IncludeMapEntry libc_symbol_map[] = {
   // It is unspecified if the cname headers provide ::size_t.
   // <locale.h> is the one header which defines NULL but not size_t.
   { "size_t", kPrivate, "<stddef.h>", kPublic },  // 'canonical' location for size_t
-  { "size_t", kPrivate, "<aio.h>", kPublic },
-  { "size_t", kPrivate, "<glob.h>", kPublic },
-  { "size_t", kPrivate, "<grp.h>", kPublic },
-  { "size_t", kPrivate, "<iconv.h>", kPublic },
-  { "size_t", kPrivate, "<monetary.h>", kPublic },
-  { "size_t", kPrivate, "<mqueue.h>", kPublic },
-  { "size_t", kPrivate, "<ndbm.h>", kPublic },
-  { "size_t", kPrivate, "<pwd.h>", kPublic },
-  { "size_t", kPrivate, "<regex.h>", kPublic },
-  { "size_t", kPrivate, "<search.h>", kPublic },
   { "size_t", kPrivate, "<signal.h>", kPublic },
   { "size_t", kPrivate, "<stdio.h>", kPublic },
   { "size_t", kPrivate, "<stdlib.h>", kPublic },
   { "size_t", kPrivate, "<string.h>", kPublic },
-  { "size_t", kPrivate, "<strings.h>", kPublic },
-  { "size_t", kPrivate, "<sys/mman.h>", kPublic },
-  { "size_t", kPrivate, "<sys/msg.h>", kPublic },
-  { "size_t", kPrivate, "<sys/sem.h>", kPublic },
-  { "size_t", kPrivate, "<sys/shm.h>", kPublic },
-  { "size_t", kPrivate, "<sys/socket.h>", kPublic },
   { "size_t", kPrivate, "<sys/types.h>", kPublic },
-  { "size_t", kPrivate, "<sys/uio.h>", kPublic },
   { "size_t", kPrivate, "<time.h>", kPublic },
   { "size_t", kPrivate, "<uchar.h>", kPublic },
   { "size_t", kPrivate, "<unistd.h>", kPublic },
   { "size_t", kPrivate, "<wchar.h>", kPublic },
-  { "size_t", kPrivate, "<wordexp.h>", kPublic },
   // Macros that can be defined in more than one file, don't have the
   // same __foo_defined guard that other types do, so the grep above
   // doesn't discover them.  Until I figure out a better way, I just
   // add them in by hand as I discover them.
   { "EOF", kPrivate, "<stdio.h>", kPublic },
-  { "EOF", kPrivate, "<libio.h>", kPublic },
   { "FILE", kPrivate, "<stdio.h>", kPublic },
+  { "IBSHIFT", kPrivate, "<asm/termbits.h>", kPublic },
+  { "MAP_POPULATE", kPrivate, "<sys/mman.h>", kPublic },
+  { "MAP_POPULATE", kPrivate, "<linux/mman.h>", kPublic },
   { "MAP_STACK", kPrivate, "<sys/mman.h>", kPublic },
   { "MAP_STACK", kPrivate, "<linux/mman.h>", kPublic },
+  { "MAXHOSTNAMELEN", kPrivate, "<sys/param.h>", kPublic },
+  { "MAXHOSTNAMELEN", kPrivate, "<protocols/timed.h>", kPublic },
+  { "SIGABRT", kPrivate, "<signal.h>", kPublic },
   { "SIGCHLD", kPrivate, "<signal.h>", kPublic },
-  { "SIGCHLD", kPrivate, "<linux/signal.h>", kPublic },
   { "va_list", kPrivate, "<stdarg.h>", kPublic },
   { "va_list", kPrivate, "<stdio.h>", kPublic },
   { "va_list", kPrivate, "<wchar.h>", kPublic },
@@ -316,12 +233,14 @@ const IncludeMapEntry libc_symbol_map[] = {
   { "NULL", kPrivate, "<stdlib.h>", kPublic },
   { "NULL", kPrivate, "<string.h>", kPublic },
   { "NULL", kPrivate, "<time.h>", kPublic },
+  { "NULL", kPrivate, "<unistd.h>", kPublic },
   { "NULL", kPrivate, "<wchar.h>", kPublic },
+  { "offsetof", kPrivate, "<stddef.h>", kPublic },
 };
 
-// Symbol -> include mappings for GNU libstdc++
-const IncludeMapEntry libstdcpp_symbol_map[] = {
-  // Kludge time: almost all STL types take an allocator, but they
+// Common kludges for C++ standard libraries
+const IncludeMapEntry stdlib_cxx_symbol_map[] = {
+  // Almost all STL types take an allocator, but they
   // almost always use the default value.  Usually we detect that
   // and don't try to do IWYU, but sometimes it passes through.
   // For instance, when adding two strings, we end up calling
@@ -354,6 +273,12 @@ const IncludeMapEntry libstdcpp_symbol_map[] = {
   { "std::size_t", kPrivate, "<cwchar>", kPublic },
 };
 
+// Symbol -> include mappings for GNU libstdc++
+const IncludeMapEntry libstdcpp_symbol_map[] = {
+  // GCC defines std::declval in <type_traits>, but the canonical location is <utility>
+  { "std::declval", kPrivate, "<utility>", kPublic },
+};
+
 const IncludeMapEntry libc_include_map[] = {
   // Private -> public include mappings for GNU libc
   // ( cd /usr/include && grep '^ *# *include' {sys/,net/,}* | perl -nle 'm/^([^:]+).*<([^>]+)>/ && print qq@    { "<$2>", kPrivate, "<$1>", kPublic },@' | grep bits/ | sort )
@@ -395,6 +320,7 @@ const IncludeMapEntry libc_include_map[] = {
   { "<bits/mathdef.h>", kPrivate, "<math.h>", kPublic },
   { "<bits/mathinline.h>", kPrivate, "<math.h>", kPublic },
   { "<bits/mman.h>", kPrivate, "<sys/mman.h>", kPublic },
+  { "<bits/mman-shared.h>", kPrivate, "<sys/mman.h>", kPublic },
   { "<bits/monetary-ldbl.h>", kPrivate, "<monetary.h>", kPublic },
   { "<bits/mqueue.h>", kPrivate, "<mqueue.h>", kPublic },
   { "<bits/mqueue2.h>", kPrivate, "<mqueue.h>", kPublic },
@@ -446,18 +372,25 @@ const IncludeMapEntry libc_include_map[] = {
   { "<bits/string2.h>", kPrivate, "<string.h>", kPublic },
   { "<bits/string3.h>", kPrivate, "<string.h>", kPublic },
   { "<bits/stropts.h>", kPrivate, "<stropts.h>", kPublic },
+  { "<bits/struct_stat.h>", kPrivate, "<sys/stat.h>", kPublic },
+  { "<bits/struct_stat.h>", kPrivate, "<ftw.h>", kPublic },
   { "<bits/sys_errlist.h>", kPrivate, "<stdio.h>", kPublic },
-  { "<bits/syscall.h>", kPrivate, "<sys/syscall.h>", kPrivate },
+  { "<bits/syscall.h>", kPrivate, "<sys/syscall.h>", kPublic },
   { "<bits/sysctl.h>", kPrivate, "<sys/sysctl.h>", kPublic },
   { "<bits/syslog-ldbl.h>", kPrivate, "<sys/syslog.h>", kPrivate },
   { "<bits/syslog-path.h>", kPrivate, "<sys/syslog.h>", kPrivate },
   { "<bits/syslog.h>", kPrivate, "<sys/syslog.h>", kPrivate },
+  { "<bits/termios-c_lflag.h>", kPrivate, "<termios.h>", kPublic },
+  { "<bits/termios-struct.h>", kPrivate, "<termios.h>", kPublic },
+  { "<bits/termios-tcflow.h>", kPrivate, "<termios.h>", kPublic },
   { "<bits/termios.h>", kPrivate, "<termios.h>", kPublic },
   { "<bits/time.h>", kPrivate, "<time.h>", kPublic },
   { "<bits/time.h>", kPrivate, "<sys/time.h>", kPublic },
   { "<bits/timerfd.h>", kPrivate, "<sys/timerfd.h>", kPublic },
   { "<bits/timex.h>", kPrivate, "<sys/timex.h>", kPublic },
   { "<bits/types.h>", kPrivate, "<sys/types.h>", kPublic },
+  { "<bits/types/siginfo_t.h>", kPrivate, "<signal.h>", kPublic },
+  { "<bits/types/siginfo_t.h>", kPrivate, "<sys/wait.h>", kPublic },
   { "<bits/uio.h>", kPrivate, "<sys/uio.h>", kPublic },
   { "<bits/unistd.h>", kPrivate, "<unistd.h>", kPublic },
   { "<bits/ustat.h>", kPrivate, "<sys/ustat.h>", kPrivate },
@@ -522,7 +455,6 @@ const IncludeMapEntry libc_include_map[] = {
   { "<bits/string.h>", kPrivate, "<string.h>", kPublic },
   { "<bits/string2.h>", kPrivate, "<string.h>", kPublic },
   { "<bits/string3.h>", kPrivate, "<string.h>", kPublic },
-  { "<bits/syscall.h>", kPrivate, "<sys/syscall.h>", kPrivate },
   { "<bits/timerfd.h>", kPrivate, "<sys/timerfd.h>", kPublic },
   { "<bits/typesizes.h>", kPrivate, "<sys/types.h>", kPublic },
   // Top-level #includes that just forward to another file:
@@ -533,7 +465,6 @@ const IncludeMapEntry libc_include_map[] = {
   // to decide which of the two files is canonical.  If neither is
   // on the POSIX.1 1998 list, I just choose the top-level one.
   { "<sys/poll.h>", kPrivate, "<poll.h>", kPublic },
-  { "<sys/syscall.h>", kPrivate, "<syscall.h>", kPublic },
   { "<sys/syslog.h>", kPrivate, "<syslog.h>", kPublic },
   { "<sys/ustat.h>", kPrivate, "<ustat.h>", kPublic },
   { "<wait.h>", kPrivate, "<sys/wait.h>", kPublic },
@@ -555,7 +486,7 @@ const IncludeMapEntry libc_include_map[] = {
   { "<asm/errno.h>", kPrivate, "<errno.h>", kPublic },
   { "<asm/errno-base.h>", kPrivate, "<errno.h>", kPublic },
   { "<asm/ptrace-abi.h>", kPrivate, "<asm/ptrace.h>", kPublic },
-  { "<asm/unistd.h>", kPrivate, "<syscall.h>", kPublic },
+  { "<asm/unistd.h>", kPrivate, "<sys/syscall.h>", kPublic },
   { "<linux/limits.h>", kPrivate, "<limits.h>", kPublic },   // PATH_MAX
   { "<linux/prctl.h>", kPrivate, "<sys/prctl.h>", kPublic },
   { "<sys/ucontext.h>", kPrivate, "<ucontext.h>", kPublic },
@@ -691,6 +622,31 @@ const char* stdlib_cpp_public_headers[] = {
   "<version>",
 };
 
+// Common public -> public include mappings for C++ standard library
+//
+// Note: make sure to sync this setting with stl.public.imp
+//
+const IncludeMapEntry stdlib_cpp_include_map[] = {
+  // The iostream .h files are confusing.  Lots of private headers,
+  // which are handled below, but we also have public headers
+  // #including each other (eg <iostream> #includes <istream>).  We
+  // are pretty forgiving: if a user specifies any public header, we
+  // generally don't require the others.
+  // ( cd /usr/crosstool/v12/gcc-4.3.1-glibc-2.3.6-grte/x86_64-unknown-linux-gnu/x86_64-unknown-linux-gnu/include/c++/4.3.1 && egrep '^ *# *include <(istream|ostream|iostream|fstream|sstream|streambuf|ios|iosfwd)>' *stream* ios | perl -nle 'm/^([^:]+).*[<"]([^>"]+)[>"]/ and print qq@    { "<$2>", kPublic, "<$1>", kPublic },@' | sort -u )
+  { "<ios>", kPublic, "<istream>", kPublic },
+  { "<ios>", kPublic, "<ostream>", kPublic },
+  { "<iosfwd>", kPublic, "<ios>", kPublic },
+  { "<iosfwd>", kPublic, "<streambuf>", kPublic },
+  { "<istream>", kPublic, "<fstream>", kPublic },
+  { "<istream>", kPublic, "<iostream>", kPublic },
+  { "<istream>", kPublic, "<sstream>", kPublic },
+  { "<ostream>", kPublic, "<fstream>", kPublic },
+  { "<ostream>", kPublic, "<iostream>", kPublic },
+  { "<ostream>", kPublic, "<istream>", kPublic },
+  { "<ostream>", kPublic, "<sstream>", kPublic },
+  { "<streambuf>", kPublic, "<ios>", kPublic },
+};
+
 // Private -> public include mappings for GNU libstdc++
 //
 // Note: make sure to sync this setting with gcc.stl.headers.imp
@@ -821,6 +777,7 @@ const IncludeMapEntry libstdcpp_include_map[] = {
   { "<bits/unique_ptr.h>", kPrivate, "<memory>", kPublic },
   { "<bits/unordered_map.h>", kPrivate, "<unordered_map>", kPublic },
   { "<bits/unordered_set.h>", kPrivate, "<unordered_set>", kPublic },
+  { "<bits/utility.h>", kPrivate, "<utility>", kPublic },
   { "<bits/valarray_after.h>", kPrivate, "<valarray>", kPublic },
   { "<bits/valarray_array.h>", kPrivate, "<valarray>", kPublic },
   { "<bits/valarray_array.tcc>", kPrivate, "<valarray>", kPublic },
@@ -985,24 +942,6 @@ const IncludeMapEntry libstdcpp_include_map[] = {
   { "<hashtable.h>", kPrivate, "<hash_set>", kPublic },
   // (This one should perhaps be found automatically somehow.)
   { "<ext/sso_string_base.h>", kPrivate, "<string>", kPublic },
-  // The iostream .h files are confusing.  Lots of private headers,
-  // which are handled above, but we also have public headers
-  // #including each other (eg <iostream> #includes <istream>).  We
-  // are pretty forgiving: if a user specifies any public header, we
-  // generally don't require the others.
-  // ( cd /usr/crosstool/v12/gcc-4.3.1-glibc-2.3.6-grte/x86_64-unknown-linux-gnu/x86_64-unknown-linux-gnu/include/c++/4.3.1 && egrep '^ *# *include <(istream|ostream|iostream|fstream|sstream|streambuf|ios|iosfwd)>' *stream* ios | perl -nle 'm/^([^:]+).*[<"]([^>"]+)[>"]/ and print qq@    { "<$2>", kPublic, "<$1>", kPublic },@' | sort -u )
-  { "<ios>", kPublic, "<istream>", kPublic },
-  { "<ios>", kPublic, "<ostream>", kPublic },
-  { "<iosfwd>", kPublic, "<ios>", kPublic },
-  { "<iosfwd>", kPublic, "<streambuf>", kPublic },
-  { "<istream>", kPublic, "<fstream>", kPublic },
-  { "<istream>", kPublic, "<iostream>", kPublic },
-  { "<istream>", kPublic, "<sstream>", kPublic },
-  { "<ostream>", kPublic, "<fstream>", kPublic },
-  { "<ostream>", kPublic, "<iostream>", kPublic },
-  { "<ostream>", kPublic, "<istream>", kPublic },
-  { "<ostream>", kPublic, "<sstream>", kPublic },
-  { "<streambuf>", kPublic, "<ios>", kPublic },
   // The location of exception_defines.h varies by GCC version.  It should
   // never be included directly.
   { "<exception_defines.h>", kPrivate, "<exception>", kPublic },
@@ -1016,6 +955,56 @@ const IncludeMapEntry libstdcpp_include_map[] = {
   { "<tbb/tbb_stddef.h>", kPrivate, "<execution>", kPublic },
 };
 
+const IncludeMapEntry libcxx_symbol_map[] = {
+    {"std::nullptr_t", kPrivate, "<cstddef>", kPublic},
+
+    // For older MacOS libc++ (13.0.0), on macOS Ventura (13.2.1)
+    {"std::string", kPrivate, "<string>", kPublic},
+};
+
+const IncludeMapEntry libcxx_include_map[] = {
+    {"<__mutex_base>", kPrivate, "<mutex>", kPublic},
+    {"<__tree>", kPrivate, "<set>", kPublic},
+    {"<__tree>", kPrivate, "<map>", kPublic},
+
+    // For the following entries:
+    // cd llvm-project/libcxx/include ; find -type d -name "__*" | sort | sed -e "s#./__\(.*\)#  { \"@<__\1/.*>\", kPrivate, \"<\1>\", kPublic },#"
+    //
+    // tweak tuple_dir entry, and comment out debug_utils, fwd, support
+    {"@<__algorithm/.*>", kPrivate, "<algorithm>", kPublic},
+    {"@<__atomic/.*>", kPrivate, "<atomic>", kPublic},
+    {"@<__algorithm/.*>", kPrivate, "<algorithm>", kPublic},
+    {"@<__bit/.*>", kPrivate, "<bit>", kPublic},
+    {"@<__charconv/.*>", kPrivate, "<charconv>", kPublic},
+    {"@<__chrono/.*>", kPrivate, "<chrono>", kPublic},
+    {"@<__compare/.*>", kPrivate, "<compare>", kPublic},
+    {"@<__concepts/.*>", kPrivate, "<concepts>", kPublic},
+    {"@<__coroutine/.*>", kPrivate, "<coroutine>", kPublic},
+    //{ "@<__debug_utils/.*>", kPrivate, "<debug_utils>", kPublic },
+    {"@<__expected/.*>", kPrivate, "<expected>", kPublic},
+    {"@<__filesystem/.*>", kPrivate, "<filesystem>", kPublic},
+    {"@<__format/.*>", kPrivate, "<format>", kPublic},
+    {"@<__functional/.*>", kPrivate, "<functional>", kPublic},
+    //{ "@<__fwd/.*>", kPrivate, "<fwd>", kPublic },
+    {"@<__ios/.*>", kPrivate, "<ios>", kPublic},
+    {"@<__iterator/.*>", kPrivate, "<iterator>", kPublic},
+    {"@<__memory/.*>", kPrivate, "<memory>", kPublic},
+    {"@<__memory_resource/.*>", kPrivate, "<memory_resource>", kPublic},
+    {"@<__numeric/.*>", kPrivate, "<numeric>", kPublic},
+    {"@<__random/.*>", kPrivate, "<random>", kPublic},
+    {"@<__ranges/.*>", kPrivate, "<ranges>", kPublic},
+    {"@<__string/.*>", kPrivate, "<string>", kPublic},
+    //{ "@<__support/.*>", kPrivate, "<support>", kPublic },
+    {"@<__thread/.*>", kPrivate, "<thread>", kPublic},
+    {"@<__tuple_dir/.*>", kPrivate, "<tuple>", kPublic},
+    {"@<__type_traits/.*>", kPrivate, "<type_traits>", kPublic},
+    {"@<__utility/.*>", kPrivate, "<utility>", kPublic},
+    {"@<__variant/.*>", kPrivate, "<variant>", kPublic},
+
+    // For older MacOS libc++ (13.0.0), on macOS Ventura (13.2.1)
+    {"<__functional_base>", kPrivate, "<functional>", kPublic},
+};
+
 // Returns true if str is a valid quoted filepath pattern (i.e. either
 // a quoted filepath or "@" followed by a regex for matching a quoted
 // filepath).
@@ -1065,20 +1054,20 @@ void MakeNodeTransitive(IncludePicker::IncludeMap* filename_map,
                         const string& key) {
   // If we've already calculated this node's transitive closure, we're done.
   const TransitiveStatus status = (*seen_nodes)[key];
-  if (status == kCalculating) {   // means there's a cycle in the mapping
-    // TODO: Reconsider cycle handling; the include_cycle test fails without
-    // this special-casing, but it seems we should handle this more generally.
-    if (key.find("internal/") != string::npos) {
-      VERRS(4) << "Ignoring a cyclical mapping involving " << key << "\n";
-      return;
-    }
-  }
-  if (status == kCalculating) {
-    VERRS(0) << "Cycle in include-mapping:\n";
+  if (status == kCalculating) {  // means there's a cycle in the mapping
+    // Note that cycles in mappings are generally benign, the cycle detection
+    // here is only necessary to protect the recursive algorithm from infinite
+    // regress. We will still expand all reachable nodes in the graph to a
+    // plain sequence representing the transitive closure.
+    // The expanded mappings are only used for simple lookup, never followed
+    // recursively (which could have necessitated preserving cycles and handling
+    // them in that traversal too).
+    // Log cycles at a high verbosity level to aid debugging.
+    VERRS(8) << "Ignored cycle in include mappings: ";
     for (const string& node : *node_stack)
-      VERRS(0) << "  " << node << " ->\n";
-    VERRS(0) << "  " << key << "\n";
-    CHECK_UNREACHABLE_("Cycle in include-mapping");  // cycle is a fatal error
+      VERRS(8) << node << " -> ";
+    VERRS(8) << key << "\n";
+    return;
   }
   if (status == kDone)
     return;
@@ -1204,26 +1193,74 @@ bool MappedInclude::HasAbsoluteQuotedInclude() const {
   return IsAbsolutePath(path);
 }
 
-IncludePicker::IncludePicker(bool no_default_mappings)
-    : has_called_finalize_added_include_lines_(false) {
-  if (!no_default_mappings) {
-    AddDefaultMappings();
-  }
+IncludePicker::IncludePicker(RegexDialect regex_dialect,
+                             CStdLib cstdlib,
+                             CXXStdLib cxxstdlib)
+    : has_called_finalize_added_include_lines_(false),
+      regex_dialect(regex_dialect) {
+  AddDefaultMappings(cstdlib, cxxstdlib);
 }
 
-void IncludePicker::AddDefaultMappings() {
-  AddSymbolMappings(libc_symbol_map, IWYU_ARRAYSIZE(libc_symbol_map));
-  AddSymbolMappings(libstdcpp_symbol_map, IWYU_ARRAYSIZE(libstdcpp_symbol_map));
+void IncludePicker::AddDefaultMappings(CStdLib cstdlib,
+                                       CXXStdLib cxxstdlib) {
+  using clang::tooling::stdlib::Header;
+  using clang::tooling::stdlib::Lang;
+  using clang::tooling::stdlib::Symbol;
 
-  AddIncludeMappings(libc_include_map,
-      IWYU_ARRAYSIZE(libc_include_map));
-  AddIncludeMappings(stdlib_c_include_map,
-      IWYU_ARRAYSIZE(stdlib_c_include_map));
-  AddIncludeMappings(libstdcpp_include_map,
-      IWYU_ARRAYSIZE(libstdcpp_include_map));
+  if (cstdlib == CStdLib::Glibc) {
+    AddSymbolMappings(libc_symbol_map, IWYU_ARRAYSIZE(libc_symbol_map));
+    AddIncludeMappings(libc_include_map, IWYU_ARRAYSIZE(libc_include_map));
+  } else if (cstdlib == CStdLib::ClangSymbols) {
+    // Get canonical C standard library mappings from clang tooling
+    for (const Symbol& sym : Symbol::all(Lang::C)) {
+      string name = sym.name().str();
 
-  AddPublicIncludes(stdlib_cpp_public_headers,
-      IWYU_ARRAYSIZE(stdlib_cpp_public_headers));
+      // the canonical header is returned first
+      for (const Header& header : sym.headers()) {
+        AddSymbolMapping(name, MappedInclude(header.name().str()), kPublic);
+      }
+    }
+  }
+
+  if (cxxstdlib == CXXStdLib::Libstdcxx) {
+    AddSymbolMappings(libstdcpp_symbol_map,
+                      IWYU_ARRAYSIZE(libstdcpp_symbol_map));
+    AddIncludeMappings(libstdcpp_include_map,
+                       IWYU_ARRAYSIZE(libstdcpp_include_map));
+  } else if (cxxstdlib == CXXStdLib::Libcxx) {
+    AddSymbolMappings(libcxx_symbol_map, IWYU_ARRAYSIZE(libcxx_symbol_map));
+    AddIncludeMappings(libcxx_include_map, IWYU_ARRAYSIZE(libcxx_include_map));
+  } else if (cxxstdlib == CXXStdLib::ClangSymbols) {
+    // Get canonical C++ standard library mappings from clang tooling
+    for (const Symbol& sym : Symbol::all(Lang::CXX)) {
+      string name = sym.qualifiedName().str();
+
+      // the canonical header is returned first
+      for (const Header& header : sym.headers()) {
+        AddSymbolMapping(name, MappedInclude(header.name().str()), kPublic);
+      }
+    }
+  }
+
+  if (cxxstdlib != CXXStdLib::None) {
+    // Map C headers to associated C++ headers. The standard library
+    // mappings shouldn't be mentioning the C headers.
+    AddIncludeMappings(stdlib_c_include_map,
+                       IWYU_ARRAYSIZE(stdlib_c_include_map));
+    // C++ include mappings to allow different public headers that
+    // generally include each other.
+    AddIncludeMappings(stdlib_cpp_include_map,
+                       IWYU_ARRAYSIZE(stdlib_cpp_include_map));
+
+    // Add common C++ mappings to deal with generic C++ standard
+    // library symbol issues (so the standard library doesn't have to
+    // do this too). If it does that's ok.
+    AddSymbolMappings(stdlib_cxx_symbol_map,
+                      IWYU_ARRAYSIZE(stdlib_cxx_symbol_map));
+
+    AddPublicIncludes(stdlib_cpp_public_headers,
+                      IWYU_ARRAYSIZE(stdlib_cpp_public_headers));
+  }
 }
 
 void IncludePicker::MarkVisibility(VisibilityMap* map,
@@ -1277,6 +1314,7 @@ void IncludePicker::AddDirectInclude(const string& includer_filepath,
     // the closing quote as part of the .*.
     AddFriendRegex(includee_filepath,
                    quoted_includee.substr(0, internal_pos) + ".*");
+    VERRS(8) << "Adding dynamic mapping for internal/ header\n";
     AddMapping(quoted_includee, mapped_includer);
   }
 
@@ -1286,6 +1324,7 @@ void IncludePicker::AddDirectInclude(const string& includer_filepath,
     string public_header = quoted_includee;
     StripPast(&public_header, "/");   // read past "asm-whatever/"
     public_header = "<asm/" + public_header;   // now it's <asm/something.h>
+    VERRS(8) << "Adding dynamic mapping for <asm-*> header\n";
     AddMapping(quoted_includee, MappedInclude(public_header));
   }
 }
@@ -1407,19 +1446,20 @@ void IncludePicker::ExpandRegexes() {
   for (const auto& incmap : quoted_includes_to_quoted_includers_) {
     const string& hdr = incmap.first;
     for (const string& regex_key : filepath_include_map_regex_keys) {
+      const string regex = regex_key.substr(1);
       const vector<MappedInclude>& map_to = filepath_include_map_[regex_key];
-      // Enclose the regex in ^(...)$ for full match.
-      std::regex regex(std::string("^(" + regex_key.substr(1) + ")$"));
-      if (std::regex_match(hdr, regex) &&
+      if (RegexMatch(regex_dialect, hdr, regex) &&
           !ContainsQuotedInclude(map_to, hdr)) {
-        Extend(&filepath_include_map_[hdr], filepath_include_map_[regex_key]);
+        for (const MappedInclude& target : map_to) {
+          filepath_include_map_[hdr].push_back(MappedInclude(
+              RegexReplace(regex_dialect, hdr, regex, target.quoted_include)));
+        }
         MarkVisibility(&include_visibility_map_, hdr,
                        include_visibility_map_[regex_key]);
       }
     }
     for (const string& regex_key : friend_to_headers_map_regex_keys) {
-      std::regex regex(std::string("^(" + regex_key.substr(1) + ")$"));
-      if (std::regex_match(hdr, regex)) {
+      if (RegexMatch(regex_dialect, hdr, regex_key.substr(1))) {
         InsertAllInto(friend_to_headers_map_[regex_key],
                       &friend_to_headers_map_[hdr]);
       }
