@@ -1,4 +1,4 @@
-#!/usr/bin/env python
+#!/usr/bin/env python3
 
 ##===--- iwyu_test_util.py - include-what-you-use test framework ----------===##
 #
@@ -10,16 +10,6 @@
 ##===----------------------------------------------------------------------===##
 
 """Utilities for writing tests for IWYU.
-
-This script has been tested with python 2.7, 3.1.3 and 3.2.
-In order to support all of these platforms there are a few unusual constructs:
- * print statements require parentheses
- * standard output must be decoded as utf-8
- * range() must be used in place of xrange()
- * _PortableNext() is used to obtain next iterator value
-
-There is more detail on some of these issues at:
-http://diveintopython3.org/porting-code-to-python-3-with-2to3.html
 """
 
 __author__ = 'wan@google.com (Zhanyong Wan)'
@@ -28,8 +18,10 @@ import difflib
 import operator
 import os
 import re
+import shlex
 import subprocess
 import sys
+import unittest
 
 # These are the warning/error lines that iwyu.cc produces when --verbose >= 3
 _EXPECTED_DIAGNOSTICS_RE = re.compile(r'^\s*//\s*IWYU:\s*(.*)$')
@@ -54,12 +46,54 @@ _NODIFFS_RE = re.compile(r'^\((.*?) has correct #includes/fwd-decls\)$')
 # // IWYU_ARGS: -Xiwyu --mapping_file=... -I .
 _IWYU_TEST_RUN_ARGS_RE = re.compile(r'^//\sIWYU_ARGS:\s(.*)$')
 
+# Text matching a condition, either as part of IWYU_REQUIRES,
+# IWYU_UNSUPPORTED or IWYU_XFAIL.
+_IWYU_CONDITION = r'([a-z][a-z0-9_-]*)\(([^)]+)\)'
 
-def _PortableNext(iterator):
-  if hasattr(iterator, 'next'):
-    iterator.next()  # Python 2.4-2.6
-  else:
-    next(iterator)   # Python 3
+# This is a line that specifies prerequisites for a test (test will be skipped
+# if they don't all hold). IWYU_REQUIRES specifies a condition that must be true
+# for the test to execute. IWYU_UNSUPPORTED specifies a condition in which the
+# test is invalid, and cannot be executed. Multiple IWYU_REQUIRES and
+# IWYU_UNSUPPORTED lines can be specified. Examples:
+# // IWYU_REQUIRES: feature(arg)
+# // IWYU_UNSUPPORTED: feature(arg)
+_IWYU_TEST_PREREQ_RE = re.compile(r'^// IWYU_(REQUIRES|UNSUPPORTED): ' +
+                                  _IWYU_CONDITION + r'$')
+
+# This is a line that specifies conditions under which a test is expected to
+# fail.
+# // IWYU_XFAIL: feature(arg)
+_IWYU_TEST_XFAIL_RE = re.compile(r'^// IWYU_XFAIL(?:: ' +
+                                 _IWYU_CONDITION + r')?$')
+
+class Features:
+  """ Implements feature support functions for prerequisites and
+      expected failures. """
+
+  # Parses the output of 'include-what-you-use -print-targets'.
+  SUPPORTED_TARGET_RE = re.compile(r'^\s+([a-z0-9_-]+)\s* - .*$')
+
+  def __init__(self):
+    # Maps all known features names to prerequisite predicates.
+    self.known = {
+      'has-target': Features.HasTarget
+    }
+
+  @staticmethod
+  def HasTarget(target):
+    iwyu = _GetIwyuPath()
+    result = subprocess.run([iwyu, '-print-targets'],
+                            capture_output=True,
+                            text=True)
+    supported = []
+    for line in result.stdout.splitlines():
+      m = Features.SUPPORTED_TARGET_RE.match(line)
+      if m:
+        supported.append(m.group(1))
+    return target in supported
+
+
+_FEATURES = Features()
 
 
 def _Which(program, paths):
@@ -117,15 +151,8 @@ def _GetIwyuPath():
   return _IWYU_PATH
 
 
-def _ShellQuote(arg):
-  if ' ' in arg:
-    arg = '"' + arg + '"'
-  return arg
-
-
 def _GetCommandOutput(command):
   p = subprocess.Popen(command,
-                       shell=True,
                        stdout=subprocess.PIPE,
                        stderr=subprocess.STDOUT)
   stdout, _ = p.communicate()
@@ -136,7 +163,7 @@ def _GetCommandOutput(command):
 
 def _GetMatchingLines(regex, file_names):
   """Returns a map: file location => string matching `regex`.
-  
+
   File location is a tuple (file_name, line number starting from 1)."""
 
   loc_to_line = {}
@@ -406,7 +433,7 @@ def _CompareExpectedAndActualSummaries(expected_summaries, actual_summaries):
     this_failure = difflib.unified_diff(expected_summaries.get(loc, []),
                                         actual_summaries.get(loc, []))
     try:
-      _PortableNext(this_failure)     # read past the 'what files are this' header
+      next(this_failure)     # read past the 'what files are this' header
       failures.append('\n')
       failures.append('Unexpected summary diffs for %s:\n' % loc)
       failures.extend(this_failure)
@@ -436,10 +463,87 @@ def _GetLaunchArguments(cc_file):
       args = args[:-1] + ' ' + line
 
     if args.endswith('\\'):
-      raise SyntaxError('%s:%s syntax error in multiline IWYU_ARGS' % 
+      raise SyntaxError('%s:%s syntax error in multiline IWYU_ARGS' %
           (cc_file, lineno))
 
-  return args
+  return shlex.split(args)
+
+
+def _ParsePrerequisites(cc_file):
+  """ Parses test prerequisites out of cc_file. """
+  prerequisites = []
+  with open(cc_file) as it:
+    for lineno, line in enumerate(it):
+      m = _IWYU_TEST_PREREQ_RE.match(line)
+      if not m:
+        continue
+
+      directive = m.group(1)
+      feature = m.group(2)
+      value = m.group(3)
+
+      feature_pred = _FEATURES.known.get(feature)
+      if not feature_pred:
+        raise SyntaxError('%s:%s IWYU_%s: unsupported feature: %s' %
+                          (cc_file, lineno, directive, feature))
+
+      if directive == 'REQUIRES':
+        prerequisite = lambda arg=value: feature_pred(arg)
+      elif directive == 'UNSUPPORTED':
+        prerequisite = lambda arg=value: not feature_pred(arg)
+      else:
+        assert False, ('_IWYU_TEST_PREREQ_RE should not have matched: %s' %
+                       directive)
+
+      desc = '%s: %s(%s)' % (directive, feature, value)
+      prerequisites.append((prerequisite, desc))
+    return prerequisites
+
+
+def _CheckPrerequisites(cc_file):
+  """ Raises a SkipTest exception if any prerequisites in cc_file fail. """
+  prerequisites = _ParsePrerequisites(cc_file)
+
+  failed = []
+  for prerequisite, desc in prerequisites:
+    if not prerequisite():
+      failed.append(desc)
+
+  if failed:
+    raise unittest.SkipTest(', '.join(failed))
+
+
+def _ParseExpectedFailures(cc_file):
+  """ Parses expected failure conditions out of cc_file. """
+  conditions = []
+  with open(cc_file) as it:
+    for lineno, line in enumerate(it):
+      m = _IWYU_TEST_XFAIL_RE.match(line)
+      if not m:
+        continue
+
+      feature = m.group(1)
+      value = m.group(2)
+
+      if feature is None:
+        # Bare IWYU_XFAIL, condition is always True
+        condition = lambda: True
+      else:
+        feature_cond = _FEATURES.known.get(feature)
+        if not feature_cond:
+          raise SyntaxError('%s:%s IWYU_XFAIL: unsupported feature: %s' %
+                            (cc_file, lineno, feature))
+
+        condition = lambda arg=value: feature_cond(arg)
+
+      conditions.append(condition)
+    return conditions
+
+
+def IsTestExpectedToFail(cc_file):
+  """ Checks if a test is expected to fail. """
+  xfail_conditions = _ParseExpectedFailures(cc_file)
+  return any(condition() for condition in xfail_conditions)
 
 
 def TestIwyuOnRelativeFile(cc_file, cpp_files_to_check, verbose=False):
@@ -451,24 +555,26 @@ def TestIwyuOnRelativeFile(cc_file, cpp_files_to_check, verbose=False):
               to check the diagnostics on, relative to the current dir.
     verbose: Whether to display verbose output.
   """
-  verbosity_flags = []
+  # Parse and check IWYU_{REQUIRES,UNSUPPORTED}
+  _CheckPrerequisites(cc_file)
+
+  cmd = [_GetIwyuPath()]
+  # Require verbose level 3 so that we can verify the individual diagnostics.
+  # We allow the level to be overriden by
+  # * IWYU_ARGS comment in a test file
+  # * IWYU_VERBOSE environment variable
+  cmd += ['-Xiwyu', '--verbose=3']
+  cmd += _GetLaunchArguments(cc_file)
   env_verbose_level = os.getenv('IWYU_VERBOSE')
   if env_verbose_level:
-    verbosity_flags = ['-Xiwyu', '--verbose=' + env_verbose_level]
+    cmd += ['-Xiwyu', '--verbose=' + env_verbose_level]
+  env_iwyu_extra_args = os.getenv('IWYU_EXTRA_ARGS')
+  if env_iwyu_extra_args:
+      cmd += shlex.split(env_iwyu_extra_args)
+  cmd += [cc_file]
 
-  cmd = '%s %s %s %s %s' % (
-    _ShellQuote(_GetIwyuPath()),
-    # Require verbose level 3 so that we can verify the individual diagnostics.
-    # We allow the level to be overriden by
-    # * IWYU_ARGS comment in a test file;
-    # * iwyu_flags;
-    # * IWYU_VERBOSE environment variable;
-    '-Xiwyu --verbose=3',
-    _GetLaunchArguments(cc_file),
-    ' '.join(verbosity_flags),
-    cc_file)
   if verbose:
-    print('>>> Running %s' % cmd)
+    print('>>> Running %s' % shlex.join(cmd))
   exit_code, output = _GetCommandOutput(cmd)
   print(''.join(output))
   sys.stdout.flush()      # don't commingle this output with the failure output
