@@ -17,11 +17,26 @@
 
 class IndirectClass {
  public:
-  void Method() const { };
+  void Method() const {
+  }
   int a;
-  static void StaticMethod() { };
+  static void StaticMethod() {
+  }
   static int statica;
 };
 
+template <typename T>
+class IndirectTemplate {
+ public:
+  void Method() const {
+  }
+  int a;
+  static void StaticMethod() {
+  }
+
+ private:
+  T t;
+};
+
 
 #endif  // INCLUDE_WHAT_YOU_USE_TESTS_CXX_INDIRECT_H_
