@@ -7,27 +7,28 @@
 //
 //===----------------------------------------------------------------------===//
 
-#include <iostream>
-#include <list>
-#include <map>
-#include <memory>
-#include "subdir/indirect_subdir.h"
+#include <list>                     // Unused C++ header to trigger diagnostics
+#include <exception>                // C++ header
+#include <errno.h>                  // C header
+#include "subdir/indirect_subdir.h" // User header
 
-std::unique_ptr<IndirectSubDirClass> CreateIndirectSubDirClass();
+inline int GetBaseError() {
+  return ENOENT;
+}
+
+extern std::exception global_exception;
+extern IndirectSubDirClass global_var;
 
 /**** IWYU_SUMMARY
 
 tests/cxx/quoted_includes_first.h should add these lines:
-class IndirectSubDirClass;
 
 tests/cxx/quoted_includes_first.h should remove these lines:
-- #include "subdir/indirect_subdir.h"  // lines XX-XX
-- #include <iostream>  // lines XX-XX
 - #include <list>  // lines XX-XX
-- #include <map>  // lines XX-XX
 
 The full include-list for tests/cxx/quoted_includes_first.h:
-#include <memory>  // for unique_ptr
-class IndirectSubDirClass;
+#include "subdir/indirect_subdir.h"  // for IndirectSubDirClass
+#include <errno.h>  // for ENOENT
+#include <exception>  // for exception
 
 ***** IWYU_SUMMARY */
